from pathlib import Path

gitignore = """# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
pip-wheel-metadata/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/
venv_*/
venv-*/
"""


def write_gitignore(path: Path) -> None:
    with open(path / ".gitignore", "w") as f:
        f.write(gitignore)



from pathlib import Path


def write_setup(path: Path, package_name: str, CLI: bool) -> None:
    setup = f"""import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_requirements = LEFT_CURLY_BRACKET
    "dev": ["wheel", "black", "pytest", "mypy"],
RIGHT_CURLY_BRACKET

setuptools.setup(
    name="{package_name}",
    version="0.1.0",
    author="[MAKEPACKAGE]",
    author_email="[MAKEPACKAGE]",
    description="[MAKEPACKAGE]",
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="[MAKEPACKAGE]",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=[],
    extras_require=extras_requirements,
    python_requires='>=3.8',
"""
    if CLI:
        setup += f"""    entry_points=LEFT_CURLY_BRACKET"console_scripts": ["{package_name} = {package_name}.__main__:main"]RIGHT_CURLY_BRACKET
)
"""
    else:
        setup += f")"

    setup = replace_curly_brackets(setup)
    with open(path / "setup.py", "w") as f:
        f.write(setup)


def replace_curly_brackets(string: str):
    """
    >>> replace_curly_brackets("LEFT_CURLY_BRACKETRIGHT_CURLY_BRACKET")
    '{}'
    >>> replace_curly_brackets(f"LEFT_CURLY_BRACKET{10*2}RIGHT_CURLY_BRACKET")
    '{20}'
    """
    return string.replace("LEFT_CURLY_BRACKET", "{").replace(
        "RIGHT_CURLY_BRACKET", "}"
    )



from pathlib import Path


def write_module_init(path: Path, package_name: str) -> None:
    module_init = f"""from .{package_name} import foo, bar, baz
"""
    with open(path / "__init__.py", "w") as f:
        f.write(module_init)



import os
from easycheck import check_if_not
from pathlib import Path

from makepackage.write_README import write_README
from makepackage.write_setup import write_setup
from makepackage.write_license import write_license
from makepackage.write_gitignore import write_gitignore
from makepackage.write_CLI_main import write_CLI_main

from makepackage.write_module_init import write_module_init
from makepackage.write_module import write_module

from makepackage.write_pytest_ini import write_pytest_ini
from makepackage.write_conftest import write_conftest
from makepackage.write_tests import write_tests
from makepackage.write_tests_init import write_tests_init


class NoPackageNameError(Exception):
    pass


class IncorrectCLIArgumentError(Exception):
    pass


class FolderExistsError(Exception):
    pass


def makepackage(package_name: str, CLI: bool) -> None:
    """Create a package package_name, with or without CLI."""

    # create directories
    root_path = (Path(".") / f"{package_name}").absolute()
    make_dirs(root_path, package_name)

    # write files in the root folder
    write_setup(root_path, package_name, CLI)
    write_README(root_path, package_name, CLI)
    write_pytest_ini(root_path)
    write_license(root_path)
    write_gitignore(root_path)

    # write files in the module's folder
    module_path = root_path / f"{package_name}"
    write_module(module_path, package_name)
    write_module_init(module_path, package_name)

    # write files in the tests/ folder
    tests_path = root_path / "tests"
    write_conftest(tests_path)
    write_tests_init(tests_path)
    write_tests(tests_path, package_name)

    # write __main__, if CLI is True
    if CLI:
        write_CLI_main(module_path, package_name)


def make_dirs(root_path: pathlib.Path, package_name: str) -> None:
    """Create the directory structure."""
    check_if_not(
        root_path.exists(),
        FolderExistsError,
        message=f"Folder {root_path} already exists.",
    )
    os.mkdir(str(root_path))
    for dir in {
        "tests",
        package_name,
    }:
        this_path = root_path / dir
        check_if_not(
            this_path.exists(),
            FolderExistsError,
            message=f"Folder {dir} already exists.",
        )
        os.mkdir(str(this_path))



from pathlib import Path


def write_CLI_main(path: Path, package_name: str) -> None:
    CLI_main = f"""from {package_name} import foo, bar, baz


def main():
    print("Running all {package_name} functions via"
        " the command-line command ($ {package_name}): "
    )
    print(
        "This is foo(): ", foo(1),
        "Function bar() uses str.lower() method:", bar("Lower"),
        "Function baz() uses str.upper() method:", baz("Upper")
    )


if __name__ == "__main__":
    main()
"""
    with open(path / "__main__.py", "w") as f:
        f.write(CLI_main)



from pathlib import Path

module = '''def foo(x: int) -> int:
    """Calculate square of x.
    
    >>> foo(2)
    4
    >>> foo(100)
    10000
    """
    return x**2


def bar(string: str) -> str:
    """Use method .lower() to a string.

    Args:
        string (str): a string to be manipulated

    Returns:
        str: a manipulated string
    
    >>> bar("Whatever!")
    'whatever!'
    >>> bar("AAaa...!")
    'aaaa...!'
    """
    return string.lower()


def baz(string: str) -> str:
    """Use method .upper() to a string.

    Args:
        string (str): a string to be manipulated

    Returns:
        str: a manipulated string
    
    >>> baz("Whatever!")
    'WHATEVER!'
    >>> baz("AAaa...!")
    'AAAA...!'
    """
    return string.upper()
'''


def write_module(path: Path, package_name: str) -> None:
    with open(path / f"{package_name}.py", "w") as f:
        f.write(module)



from pathlib import Path


def write_conftest(path: Path) -> None:
    with open(path / "conftest.py", "w") as f:
        f.write(
            """import pytest

from typing import Tuple

@pytest.fixture
def strings() -> Tuple[str, str, str]:
    return (
        "Whatever string",
        "Shout Bamalama!",
        "Sing a song about statistical models.", 
    )

"""
        )



import sys
from collections import namedtuple
from easycheck import check_length, gt, lt
from makepackage import (
    NoPackageNameError,
    IncorrectCLIArgumentError,
    makepackage,
)


def main():
    package_name, CLI = _read_cli_args()
    makepackage(package_name, CLI)
    print_final_info


def _read_cli_args():
    check_length(
        sys.argv,
        1,
        NoPackageNameError,
        operator=gt,
        message="Provide package name as the first CLI argument",
    )
    check_length(
        sys.argv,
        4,
        IncorrectCLIArgumentError,
        operator=lt,
        message="Expected two CLI args: package name and (optionally) CLI",
    )
    package_name = sys.argv[1]
    try:
        CLI = True if sys.argv[2].lower() == "cli" else False
    except IndexError:
        CLI = False
    print(
        f"Creating package <{package_name}>, "
        f"with{'' if CLI else 'out'} command-line interface."
    )
    return namedtuple("Settings", "package_name CLI")(package_name, CLI)


def print_final_info(package_name: str) -> None:
    print(
        f"Package <{package_name}> has been created.\n"
        "Check out above if all the tests (both pytests and doctests) have passed - they should."
        "To finish, you need to fill in the following fields in setup.py:\n"
        "  - author\n"
        "  - author_email\n"
        "  - description (this is a short description, as the long one is taken from README)\n"
        "You need to also fill in the author in LICENSE. You can find all those fields"
        "by searching for '[MAKEPACKAGE]' in the project.\n"
        "When you're done, the package is ready to be develop.\n"
        "Check if you need all libraries from extras_require in setup.py - they will "
        "be installed in the development mode of the package (not after installing from "
        "the wheel file)."
    )


if __name__ == "__main__":
    main()



from pathlib import Path


def write_README(path: Path, package_name: str, CLI: bool) -> None:
    README = f"""# `{package_name}`: A Python package for ...


# Installation - development

Create a virtual environment, for example using `venv`:

```shell
$ python -m venv venv-{package_name}
$ source venv-{package_name}/bin/activate
$ mkdir {package_name}
$ cd {package_name}
$ python -m pip install -e .[dev]

```

Note that the last command installs a development environment, as it also installs packages needed for development, like `black` (for code formatting), `pytest` (for unit testing) and `wheel` (for creating wheel files from the package).


# Testing

Running the tests requires to run the following command in the root folder (of course in the virtual environment):

```shell
(venv-{package_name}) > pytest
```

If you use doctests in your docstrings (as `makepackage` assumes), you can run them using the following command (in the root folder):

```shell
(venv-{package_name}) > python -m doctest {package_name}/{package_name}.py
```

In a similar way, you can run doctests from any other file that contains doctests.


## Versioning

Remember to update package version once a change is made to the package and the new version is pushed to the repository. Don't forget about releases, too.


## How to build a Python package?

To build the package, you need to go to the root folder of the package and run the following command:

```shell
(venv-{package_name}) > python setup.py sdist bdist_wheel
```

Note that this assumes you have `wheel` installed in your virtual environment, and `makepackage` does this for you.

The built package is now located in the dist/ folder.


## Publishing your package in PyPi

If you want to publish it to [PyPi](https://pypi.org/), you need to install [twine](https://twine.readthedocs.io/en/latest/), create an account there, and run the following command (also in the package's root folder):

```shell
(venv-{package_name}) > twine upload dist/*
```

Nonetheless, if you first want to check what it will look like in PyPi, you can first upload the package to [a test version of PyPi](https://test.pypi.org/), that is, 

```shell
twine upload -r testpypi dist/*
```

Check if everything is fine, and if so, you're ready to publish the package to PyPi.

## Installation from PyPi

If the package is in PyPi, you can install it from there like any other Python package, that is,

```shell
pip install 
```
"""
    if CLI:
        README += f"""

# CLI app

`{package_name}` offers a command-line interface, which you can run from shell using the following command:

```shell
$ {package_name}
```
"""

    with open(path / "README.md", "w") as f:
        f.write(README)



from pathlib import Path


def write_tests(path: Path, package_name: str) -> None:
    with open(path / f"test_{package_name}.py", "w") as f:
        f.write(
            f"""import pytest
from {package_name}.{package_name} import foo, bar, baz


@pytest.mark.parametrize(
    "test_input,expected",
    [(2, 4), (3, 9), (100, 10000), (1.2, 1.44), (-1, 1), (-2, 4), (-50, 2500),
])
def test_foo(test_input, expected):
    assert foo(test_input) == expected


def test_bar_basic():
    assert bar("AAAaA") == "aaaaa"


def test_bar(strings):
    for string in strings:
        assert not any(s.isupper() for s in bar(string))


def test_baz_basic():
    assert baz("AAAaA") == "AAAAA"


def test_baz(strings):
    for string in strings:
        assert not any(s.islower() for s in baz(string))

"""
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_requirements = {
    "dev": ["wheel", "black", "pytest", "mypy", ],
}

setuptools.setup(
    name="makepackage",
    version="0.1.2",
    author="Nyggus",
    author_email="nyggus@gmail.com",
    description="Creating a structure of a simple Python package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/nyggus/makepackage",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "easycheck",
    ],
    python_requires=">=3.6",
    extras_require=extras_requirements,
    entry_points={
        "console_scripts": ["makepackage = makepackage.__main__:main"]
    },
)
